#include <mpi.h>

#include <petscmat.h>
#include <petscerror.h>
#include <petscsys.h>
#include <petscksp.h>

#include <fstream>
#include <iostream>
#include <vector>

int main(int argc, char** argv){
    //call this with arguments <path-to-matrix> <path-to-vector>

    std::vector<char *> petsc_argv;
    petsc_argv.push_back((char*)"petscwrap");
    petsc_argv.push_back((char*) "--with-debugging=1");

    char **petsc_argv_ptr = petsc_argv.data();
    int petsc_argc = petsc_argv.size();
    
    //MPI_Init(nullptr, nullptr);

    CHKERRXX(PetscInitialize(&petsc_argc, &petsc_argv_ptr, 0, 0));
    int rank = 0; MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    int size = 0; MPI_Comm_size(MPI_COMM_WORLD, &size);

    Mat A; Vec b;

    CHKERRXX(MatCreateAIJ(
      PETSC_COMM_WORLD,   //comm
      size == 2 ? 16 : 32,//m
      size == 2 ? 16 : 32,//n
      32,                 //M
      32,                 //N
      5 * 2,              //d_nz
      nullptr,            //d_nnz
      4 * 2,              //o_nz
      nullptr,            //o_nnz
      &A                  //A
    ));

    CHKERRXX(MatGetVecs(A, &b, nullptr));

    CHKERRXX(MatAssemblyBegin(A, MAT_FINAL_ASSEMBLY));
    CHKERRXX(VecAssemblyBegin(b));

    { //fill matrix from recorded data
      std::string matrix_path = std::string(argv[1]) + std::string(".") + std::to_string(rank);
      std::ifstream ifs(matrix_path);

      int matrix_height = 0, matrix_width = 0; ifs >> matrix_height >> matrix_width;
      std::cout << "Matrix size is " << matrix_height << "x" << matrix_width << "\n";

      while(ifs.good()){
        int i,j; double d;
        ifs >> i >> j >> d;
        //std::cout << "Matrix: i:" << i << ", j:" << j << "\t" <<  d << "\n";
        CHKERRXX(MatSetValue(A, i, j, d, ADD_VALUES));
      }
#include <petscksp.h>
    }

    { //fill vector from recorded data
      std::string vector_path = std::string(argv[2]) + std::string(".") + std::to_string(rank);
      std::ifstream ifs(vector_path);

      int vector_size = 0; ifs >> vector_size;
      std::cout << "Vector size is " << vector_size << "\n";

      while(ifs.good()){
        int i; double d;
        ifs >> i >> d;
        CHKERRXX(VecSetValue(b, i, d, ADD_VALUES));
      }
    }

    //solve the system
    CHKERRXX(MatAssemblyEnd(A, MAT_FINAL_ASSEMBLY));
    CHKERRXX(VecAssemblyEnd(b));

    Vec x;
    CHKERRXX(MatGetVecs(A, &x, nullptr));
    CHKERRXX(VecSet(x, 0.));

    KSP ksp;
    CHKERRXX(KSPCreate(MPI_COMM_WORLD, &ksp));
    CHKERRXX(KSPSetType(ksp, KSPCG));

    PC pc;
    CHKERRXX(KSPGetPC(ksp, &pc));
    CHKERRXX(PCSetType(pc, PCBJACOBI));

    CHKERRXX(KSPSetOperators(ksp, A, A, DIFFERENT_NONZERO_PATTERN));

    CHKERRXX(KSPSetTolerances(ksp, 1.e-10, PETSC_DEFAULT, PETSC_DEFAULT,
                              PETSC_DEFAULT));
    CHKERRXX(KSPSolve(ksp, b, x));
    {
      double residual_norm = 0;
      int n_its = 0;
      CHKERRXX(KSPGetResidualNorm(ksp, &residual_norm));
      CHKERRXX(KSPGetIterationNumber(ksp, &n_its));
      std::cout << "PETSc KSPSolve done, residual norm: " << residual_norm
                << ", it took " << n_its << " iterations.";
    }

    CHKERRXX(MatView(A, PETSC_VIEWER_STDOUT_SELF));
    CHKERRXX(VecView(x,PETSC_VIEWER_STDOUT_SELF));

    CHKERRXX(VecDestroy(&x));
    CHKERRXX(KSPDestroy(&ksp));

    CHKERRXX(MatDestroy(&A));
    CHKERRXX(VecDestroy(&b));

    MPI_Finalize();
}
